MLFLOW_DEPLOYMENTS_HEALTH_ENDPOINT = "/health"
MLFLOW_DEPLOYMENTS_CRUD_ENDPOINT_BASE = "/api/2.0/endpoints/"
MLFLOW_DEPLOYMENTS_LIMITS_BASE = "/api/2.0/endpoints/limits/"
MLFLOW_DEPLOYMENTS_ENDPOINTS_BASE = "/endpoints/"
MLFLOW_DEPLOYMENTS_QUERY_SUFFIX = "/invocations"
MLFLOW_DEPLOYMENTS_LIST_ENDPOINTS_PAGE_SIZE = 3000
